
/* Router solicitation program.  Standalone.
   Does not require link-layer indications */


/*************************************************************/
/*
Please note that either this license MUST be displayed
or a particular instance of this code may be converted to
the GNU General Public License v2 or greater.

Please see README.txt for details.
*/

/*
Copyright (c) 2005, Monash University
All rights reserved.
Contributors: Greg Daley

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.

* Neither the name of Monash University nor the names of its
  contributors may be used to endorse or promote products derived
  from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO NTLINK SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
 
*/
/*************************************************************/

/* output */
#include <stdio.h>

/* time information */
#include <time.h>
#include <sys/time.h>

/* fd reading */
#include <unistd.h>
#include <sys/select.h>

#include <errno.h>

#include <bits/types.h>

/* memset */
#include <string.h>

/* free */
#include <stdlib.h>

/* uint32_t */
#include <stdint.h>


#include "prefixlist.h"

static double tv_diff(struct timeval *tv1, struct timeval *tv2);

void print_prefixlist(struct prefixlist *p){
	struct prefix_e *tmp;
	struct timeval now; 

        tmp=p->head->next;

	gettimeofday(&now, NULL);

        if(!p)
                return; 

        printf("print_prefixlist: %p   (%d entries) %s\n",p, p->count,
			p->complete? "COMPLETE":"");

        while(tmp != p->tail){

		printf("  %4.4x:%4.4x:%4.4x:%4.4x/%d E:%lu.%6.6lu (%f) %d\n",
                	ntohs(tmp->prefix.s6_addr16[0]),
			ntohs(tmp->prefix.s6_addr16[1]),
                	ntohs(tmp->prefix.s6_addr16[2]),
			ntohs(tmp->prefix.s6_addr16[3]), tmp->plen,
			tmp->expiry.tv_sec, tmp->expiry.tv_usec,
			tv_diff(&tmp->expiry,&now), tmp->recvd_pio);
		
                tmp=tmp->next;
        }
}

void age_prefixlist(struct prefixlist *p){
	struct prefix_e *e;
	struct prefix_e *tmp;
	struct timeval now;

        tmp=p->head->next;

	gettimeofday(&now, NULL);
	
        if(!p)
                return; 

        while(tmp != p->tail){
		if(tv_diff(&tmp->expiry, &now) > 0){
			/* remaining lifetime */
			tmp=tmp->next;
			continue;
		}
		printf("age prefix: %4.4x:%4.4x:%4.4x:%4.4x e(%lu.%6.6lu) "
			"n(%lu.%6.6lu %f) \n", ntohs(tmp->prefix.s6_addr16[0]),
			ntohs(tmp->prefix.s6_addr16[1]),
			ntohs(tmp->prefix.s6_addr16[2]),
			ntohs(tmp->prefix.s6_addr16[3]), tmp->expiry.tv_sec,
			tmp->expiry.tv_usec,now.tv_sec, now.tv_usec,
			tv_diff(&tmp->expiry,&now));
		/* clear ancient prefixes */
		e=tmp; 
		tmp=e->next;
		e->prev->next=tmp;
		tmp->prev=e->prev;
		e->next=e->prev=0;
		free(e);
		p->count--;
        }
        printf("age_prefixlist: %p   (%d entries) %s _END\n",p, p->count,
			p->complete? "COMPLETE":"");

}

void clear_prefixlist(struct prefixlist *p){
	struct prefix_e *tmp;

	fprintf(stderr,"clear prefixlist %p\n", p);

	while(!empty_prefixlist(p)){
		tmp=p->head->next;
		p->head->next=tmp->next;	
		p->head->next->prev=p->head;
		fprintf(stderr,"clear prefixlist %4.4x\n",
				 ntohl(tmp->prefix.s6_addr16[0]));
		tmp->prev=tmp->next=NULL;
		free(tmp);
	}
	p->complete=0;
	p->count=0;
	/* GD: next statement is probably redundant */
	/* remove when I've got time to test change */
	initprefixlist(p);	
}


void give_prefixlist(struct prefixlist *to, struct prefixlist *from){
	/* transfer list contents from one to another	*/
	/* destroys the contents of 'to', transfers	*/
	/* pointers of 'from'				*/

	fprintf(stderr,"give prefixlist t:%p f:%p\n",to, from);

	clear_prefixlist(to);

	fprintf(stderr,"give prefixlist: after 'to' clear\n");

	if(empty_prefixlist(from)){
		return;
	}
	fprintf(stderr,"give prefixlist: non empty from\n");
	to->head->next=from->head->next;
	to->tail->prev=from->tail->prev;
	to->tail->prev->next=to->tail;	
	to->head->next->prev=to->head;
	from->head->next=from->tail;
	from->tail->prev=from->head;
	to->count=from->count;	
	if(from->complete){
		to->complete=1;
	}
	from->complete=0;
	from->count=0;
}

void initprefixlist(struct prefixlist *p){
	uint32_t allones[4] = { 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff};

	memset(p,0, sizeof (struct prefixlist));

/*	fprintf(stderr,"initprefixlist %p \n",p );*/
	p->head=&p->h;

	p->tail=&p->t;
	p->head->plen = p->tail->plen=128;
	memcpy(&p->tail->prefix, allones, sizeof(struct in6_addr));

	p->head->prev=p->head;
	p->tail->next=p->tail;
	p->tail->prev=p->head;
	p->head->next=p->tail;
	/*fprintf(stderr,"initprefixlist %p - done\n",p );*/
}


void prefix_copy(struct in6_addr *dest, struct in6_addr *src  , int len){
	int i;	
	unsigned partoctets;
	unsigned char octmask=0xff;

	partoctets=(len+7)>>3;

	for( i = 0; i < partoctets; i++){
		dest->s6_addr[i] = src->s6_addr[i];	
	}

	if(!(len % 8)){
		octmask= 0xff;
	}
	else{
		octmask = 0xff << (8 - (len %8));
	}

	fprintf(stderr, "prefix copy: l:%d, partocts:%d, om%2.2x\n",len,
			partoctets, octmask);
	
	dest->s6_addr[partoctets-1] &= octmask;
}

int prefix_cmp(struct in6_addr *p1, struct in6_addr *p2, int l1, int l2){
	int i;
	int len;
        unsigned partoctets;
        unsigned char octmask=0xff;
	/* assume pointers non null */

	len =  (l1 < l2)? l1: l2; 

	fprintf(stderr,"prefix_cmp %4.4x:%4.4x:%4.4x:%4.4x/%d "
			"%4.4x:%4.4x:%4.4x:%4.4x/%d\n",
		ntohs(p1->s6_addr16[0]), ntohs(p1->s6_addr16[1]),
		ntohs(p1->s6_addr16[2]), ntohs(p1->s6_addr16[3]),l1,
		ntohs(p2->s6_addr16[0]), ntohs(p2->s6_addr16[1]),
		ntohs(p2->s6_addr16[2]), ntohs(p2->s6_addr16[3]),l2);

	partoctets=(len+7)>>3;

	/* compare full octets in the prefix */

	for( i = 0; i < partoctets-1; i++){
		if(p1->s6_addr[i] != p2->s6_addr[i]){
			return (p1->s6_addr[i] <  p2->s6_addr[i])?  -1: 1; 
		}
	}

	if(!(len % 8)){
		octmask= 0xff;
	}
	else{
		octmask = 0xff << (8 - (len %8));
	}

	fprintf(stderr, "prefix cmp: l1:%d l2%d partocts:%d, om%2.2x\n", l1,
		l2, partoctets, octmask);

	/* compare the last possibly non-zero prefix */
	if((p1->s6_addr[partoctets-1] & octmask) != 
		(p2->s6_addr[partoctets-1] &octmask)){
		return ((p1->s6_addr[partoctets-1] & octmask) <  
				(p2->s6_addr[partoctets] & octmask))?  -1: 1; 
	}
	/* all equal until now, tiebreak with plens */

	if(l1 != l2){
		return (l1 < l2)? -1: 1;
	}
	return 0;
}

int insert_prefix(struct in6_addr *p, int l, struct timeval *exp, int pio, 
		struct prefixlist *pfxlist){
	struct prefix_e *tmp;
	struct prefix_e *ins; 

	/* XXX: failure detection is unlikely, since the RAs will be
	 still be sent in readvertised DNAOs */
	
	int ret;
	tmp=pfxlist->head;

	if(!p || !l)
		return -1; /* failure */

	fprintf(stderr,"insert_prefix (%p) %4.4x:%4.4x:%4.4x:%4.4x/%d\n",
		pfxlist, ntohs(p->s6_addr16[0]), ntohs(p->s6_addr16[1]),
		ntohs(p->s6_addr16[2]), ntohs(p->s6_addr16[3]),l);

	while((ret = prefix_cmp(&tmp->next->prefix, p, tmp->next->plen, l))< 0){
		tmp=tmp->next;
	}
	if(!ret){/* same as another prefix	*/
		 /* XXX: update details?	pi*/
		/*if(pio || !(tmp->next->recvd_pio)){*/
			tmp->next->recvd_pio = pio;
			memcpy(&tmp->next->expiry, exp, sizeof(struct timeval));
		/*}*/
		return 0;
	}

	/* add the prefix */
	if((ins = (struct prefix_e *)malloc(sizeof(struct prefix_e)))==NULL){
		return -1;
	}
	prefix_copy(&ins->prefix, p, l);
	memcpy(&ins->expiry, exp, sizeof(struct timeval));
	ins->plen=l;
	ins->recvd_pio = pio;
	ins->prev=tmp;
	ins->next=tmp->next;
	tmp->next->prev=ins;
	tmp->next=ins;
	pfxlist->count++;
	return 1;
}

void clone_prefixlist(struct prefixlist *to, struct prefixlist *from){
	/*assumes an empty prefix list in to */
	struct prefix_e *tmp;

	tmp=from->tail->prev;

	while(tmp!=from->head){
		insert_prefix(&tmp->prefix, tmp->plen, &tmp->expiry,
				tmp->recvd_pio, to);
		tmp=tmp->prev;
	}
	to->count=from->count;
}

void supercede_prefixlist(struct prefixlist *replaced, struct prefixlist *from){
	int diff;
	struct prefix_e *tmp;
        struct prefix_e *f, *r;
	struct timeval now;
	/* transfer list contents from one to another	*/
	/* 'replaced', transfers			*/
	/* pointers of 'from', leaving it empty		*/
	/* duplicates are destroyed 			*/

	if(empty_prefixlist(replaced)){
		give_prefixlist(replaced,from);
		return;
	}

	gettimeofday(&now,NULL);
	/*printf("supercede prefixlist nonempty r:%p f:%p (%lu.%6.6lu)\n", 
		replaced, from, now.tv_sec, now.tv_usec);*/

        f=from->head->next;
        r=replaced->head->next;

        while((f != from->tail) || (r != replaced->tail)){
                diff = prefix_cmp(&f->prefix, &r->prefix, f->plen, r->plen);
		
		/*printf("r  %4.4x:%4.4x:%4.4x:%4.4x/%d E:%lu.%6.6lu %d\n",
                	ntohs(r->prefix.s6_addr16[0]),
			ntohs(r->prefix.s6_addr16[1]),
                	ntohs(r->prefix.s6_addr16[2]),
			ntohs(r->prefix.s6_addr16[3]), r->plen,
			r->expiry.tv_sec, r->expiry.tv_usec,r->recvd_pio);
		printf("f  %4.4x:%4.4x:%4.4x:%4.4x/%d E:%lu.%6.6lu %d\n",
                	ntohs(f->prefix.s6_addr16[0]),
			ntohs(f->prefix.s6_addr16[1]),
                	ntohs(f->prefix.s6_addr16[2]),
			ntohs(f->prefix.s6_addr16[3]), f->plen,
			f->expiry.tv_sec, f->expiry.tv_usec,f->recvd_pio);*/


                if(!(diff)){
			/* update r->prefix's timers */	
			/*printf("supercede prefix match (%lu.%6.6lu) "
					"(%lu.%6.6lu)\n",
					r->expiry.tv_sec,r->expiry.tv_usec,
					f->expiry.tv_sec,f->expiry.tv_usec);*/

			if(!r->recvd_pio || f->recvd_pio){
				memcpy(&r->expiry,&f->expiry,
						 sizeof (struct timeval));
				r->recvd_pio = f->recvd_pio; 
			}
			
			tmp=f;
			f->prev->next=f->next;
			f->next->prev=f->prev;
			f=f->next;
			tmp->next=tmp->prev=NULL;
			free(tmp);
			r=r->next;
                }
                else if(diff < 0){
			tmp=f;
			f->prev->next=f->next;
			f->next->prev=f->prev;
			f=f->next;

			tmp->next=r;
			tmp->prev=r->prev;
			r->prev->next=tmp;
			r->prev=tmp;			
			replaced->count++;
                }
                else{ /* doesn't match element in from: eliminate */
			tmp=r;
                        r=r->next;
			tmp->prev->next=r;
			r->prev=tmp->prev;
			tmp->next=tmp->prev=NULL;
			free(tmp);
			replaced->count--;
                }
        }
	if(from->complete){

		replaced->complete=1;
	}
	from->complete=0;
	from->count=0;
}




void union_prefixlist(struct prefixlist *unionto, struct prefixlist *from){
	int diff;
	struct prefix_e *tmp;
        struct prefix_e *f, *u;
	struct timeval now;
	/* transfer list contents from one to another	*/
	/* 'unionto', transfers				*/
	/* pointers of 'from', leaving it empty		*/
	/* duplicates are destroyed 			*/
	/*printf("union prefixlist u:%p f:%p\n", unionto, from);*/

	if(empty_prefixlist(unionto)){
		give_prefixlist(unionto,from);
		return;
	}

	gettimeofday(&now,NULL);
/*	printf("union prefixlist nonempty u:%p f:%p (%lu.%6.6lu)\n", 
		unionto, from,now.tv_sec,now.tv_usec);*/

        f=from->head->next;
        u=unionto->head->next;

        while(f != from->tail){
                diff = prefix_cmp(&f->prefix, &u->prefix, f->plen, u->plen);
		
		/*printf("u  %4.4x:%4.4x:%4.4x:%4.4x/%d E:%lu.%6.6lu %d\n",
                	ntohs(u->prefix.s6_addr16[0]),
			ntohs(u->prefix.s6_addr16[1]),
                	ntohs(u->prefix.s6_addr16[2]),
			ntohs(u->prefix.s6_addr16[3]), u->plen,
			u->expiry.tv_sec, u->expiry.tv_usec,u->recvd_pio);
		printf("f  %4.4x:%4.4x:%4.4x:%4.4x/%d E:%lu.%6.6lu %d\n",
                	ntohs(f->prefix.s6_addr16[0]),
			ntohs(f->prefix.s6_addr16[1]),
                	ntohs(f->prefix.s6_addr16[2]),
			ntohs(f->prefix.s6_addr16[3]), f->plen,
			f->expiry.tv_sec, f->expiry.tv_usec,f->recvd_pio);*/


                if(!(diff)){
			/* update u->prefix's timers */	
			/*if(!u->recvd_pio || f->recvd_pio){*/
			/*printf("union prefix match (%lu.%6.6lu) (%lu.%6.6lu)\n",
					u->expiry.tv_sec,u->expiry.tv_usec,
					f->expiry.tv_sec,f->expiry.tv_usec);*/

			if(!u->recvd_pio || f->recvd_pio){
				memcpy(&u->expiry,&f->expiry,
						 sizeof (struct timeval));
				u->recvd_pio = f->recvd_pio; 
			}
			
			tmp=f;
			f->prev->next=f->next;
			f->next->prev=f->prev;
			f=f->next;
			tmp->next=tmp->prev=NULL;
			free(tmp);
			u=u->next;
                }
                else if(diff < 0){
			tmp=f;
			f->prev->next=f->next;
			f->next->prev=f->prev;
			f=f->next;

			tmp->next=u;
			tmp->prev=u->prev;
			u->prev->next=tmp;
			u->prev=tmp;			
			unionto->count++;
                }
                else{
                        u=u->next;
                }
        }
	if(from->complete){
		unionto->complete=1;
	}
	from->complete=0;
	from->count=0;
}

int does_intersect(struct prefixlist *pl1, struct prefixlist *pl2){

	int diff;
	struct prefix_e *e1, *e2;	

	fprintf(stderr,"does_intersect pl1:%p (%d) pl2:%p (%d)\n", pl1,
			pl1->count, pl2, pl2->count);

	e1=pl1->head->next;
	e2=pl2->head->next;

	while((e1 != pl1->tail) && (e2 != pl2->tail)){
		fprintf(stderr, "Compare %x:%x:%x:%x   %x:%x:%x:%x\n",
			ntohs(e1->prefix.s6_addr16[0]),
			ntohs(e1->prefix.s6_addr16[1]),
			ntohs(e1->prefix.s6_addr16[0]),
			ntohs(e1->prefix.s6_addr16[1]),
			ntohs(e2->prefix.s6_addr16[0]),
			ntohs(e2->prefix.s6_addr16[1]),
                	ntohs(e2->prefix.s6_addr16[0]),
			ntohs(e2->prefix.s6_addr16[1]));

		if(!(diff = prefix_cmp(&e1->prefix, &e2->prefix, e1->plen,
				e2->plen))){
			return 1;
		}
		if(diff < 0){
			e1=e1->next;
		}
		else{
			e2=e2->next;
		}	
	}
	
	return 0;
}



int select_landmark(struct in6_addr *newland, struct prefixlist *pl){

	if(!newland || !pl)
		return 128; /* cowardly escape */

	if(empty_prefixlist(pl)){
		memset(newland, 0, sizeof(struct in6_addr));
		return 128;
	}
	
	/* randomly select a landmark ?? */
	
	memcpy(newland, &pl->head->next->prefix, sizeof(struct in6_addr));
	
	return  pl->head->next->plen;
}


/* GD: purloined from my own code in RADVD */
static double tv_diff(struct timeval *tv1, struct timeval *tv2){
    double sdiff=0.0;
    int sign=0;

         /* reduce gap by ms diff if tv1.sec - tv2.sec negative */
    sign=((tv2->tv_sec > tv1->tv_sec)? -1 : 1);

    sdiff = ((long int)tv1->tv_sec - (long int)tv2->tv_sec);

    sdiff += sign *
                (((long int)tv1->tv_usec - (long int)tv2->tv_usec) * 0.000001);
    return sdiff;
}

